/* TODO: Abstract styles that are shared by both collapse and display templates */

.al_collapse_template,
.al_display_template {
  padding-bottom: 1rem;
}

.al_collapse_template a.al_toggle,
.al_display_template a.al_toggle {
  color: #3b842c; /* Same as daterm in Assembly Line */
  text-decoration: dashed;
  text-decoration-line: underline;
}

.al_collapse_template .collapse,
.al_display_template .collapse {
    margin-bottom: 1rem;
}

.al_collapse_template a span.pdcaretopen,
.al_display_template a span.pdcaretopen {
    display: inline;
}

.al_collapse_template a span.pdcaretclosed,
.al_display_template a span.pdcaretclosed {
    display: none;
}

.al_collapse_template a.collapsed .pdcaretopen,
.al_display_template a.collapsed .pdcaretopen {
    display: none;
}

.al_collapse_template a.collapsed .pdcaretclosed,
.al_display_template a.collapsed .pdcaretclosed {
    display: inline;
}

/* An upstream fix for Bootstrap's lack of a dark/light mode neutral style.
   See https://github.com/jhpyle/docassemble/pull/718 */
[data-bs-theme=dark] .al_collapse_template .bg-light,
[data-bs-theme=dark] .al_display_template .bg-light,
[data-bs-theme=dark] .al_display_template.bg-light {
  --bs-bg-opacity: 1;
  background-color: rgba(var(--bs-dark-rgb), var(--bs-bg-opacity)) !important;
}
